
		/*********************************************************
		** Routines used commonly in segmentation processes;
		** specifically in the vessels probing and fuzzy convergence.
		*********************************************************/

#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "globals.h"

	/*******************************************************************
	** Given a set of points in an image (indices), of a given label (L),
	** this routine thins those points down to their minimum 8-connections.
	** Erased points are given the label 0 in the image.
	*******************************************************************/
	
void ThinPixels(unsigned char	*InputImage,
		int		nrows,int ncols,
		int		*indices,int count,	/* pixels to be thinned */
		int		L)			/* label of pixels being thinned */

{
int		i,r,c,r2,c2,pass,Transitions,EdgeNeighbors,ThinningFlag;
unsigned char	*TempArray;

TempArray=(unsigned char *)calloc(nrows*ncols,sizeof(unsigned char));
for (r=0; r<nrows*ncols; r++)
  TempArray[r]=InputImage[r];
pass=0;
do
  {
  if (pass > 0)	/* copy deletions from previous iteration into tempimage */
    for (i=0; i<count; i++)
      if (!(InputImage[indices[i]]))
        TempArray[indices[i]]=0;
  ThinningFlag=0;
  pass++;
  for (i=0; i<count; i++)
    {
    if (!(InputImage[indices[i]]))
      continue;
    r=indices[i]/ncols;
    c=indices[i]%ncols;
    Transitions=0;
    if (TempArray[(r-1)*ncols+c] != L  &&  TempArray[(r-1)*ncols+c+1] == L)
      Transitions++;
    if (TempArray[(r-1)*ncols+c+1] != L  &&  TempArray[r*ncols+c+1] == L)
      Transitions++;
    if (TempArray[r*ncols+c+1] != L  &&  TempArray[(r+1)*ncols+c+1] == L)
      Transitions++;
    if (TempArray[(r+1)*ncols+c+1] != L  &&  TempArray[(r+1)*ncols+c] == L)
      Transitions++;
    if (TempArray[(r+1)*ncols+c] != L  &&  TempArray[(r+1)*ncols+c-1] == L)
      Transitions++;
    if (TempArray[(r+1)*ncols+c-1] != L  &&  TempArray[r*ncols+c-1] == L)
      Transitions++;
    if (TempArray[r*ncols+c-1] != L  &&  TempArray[(r-1)*ncols+c-1] == L)
      Transitions++;
    if (TempArray[(r-1)*ncols+c-1] != L  &&  TempArray[(r-1)*ncols+c] == L)
      Transitions++;
    if (Transitions != 1)
      continue;	/* can only have one edge to non-edge transition */
    EdgeNeighbors=0;
    for (r2=-1; r2<=1; r2++)
      for (c2=-1; c2<=1; c2++)
        if (TempArray[(r+r2)*ncols+c+c2] == L)
  	EdgeNeighbors++;
    if (EdgeNeighbors < 3  ||  EdgeNeighbors > 7)
      continue;	/* must have 3 to 7 edge pixel neighbors */
    if (TempArray[((r-1)*ncols)+c] != L  || 
		TempArray[(r*ncols)+c+1] != L  ||
		(TempArray[(r*ncols)+c-1] != L  &&
		TempArray[((r+1)*ncols)+c] != L))
      {
      ThinningFlag=1;
      InputImage[(r*ncols)+c]=0;
      }
    }
  }
while (ThinningFlag == 1);
free(TempArray);
}





	/*******************************************************************
	** Given a set of thinned points in an image (indices), of a given
	** label (L), this routine counts up the total branches and endpoints
	** of the set.  The indices and the counts are returned.
	*******************************************************************/

void BranchesAndEnds(unsigned char	*InputImage,
		int		nrows,int ncols,
		int		*indices,int total_indices,	/* indices of points to check */
		int		*endpoints,int *total_ends,		/* returned */
		int		*branchpoints,int *total_branches,	/* returned */
		int		max_points,	  /* size of arrays for holding points */
		int		L)		/* label of region to find endpoints for */

{
int		i,r,c,Transitions;

(*total_ends)=(*total_branches)=0;
for (i=0; i<total_indices; i++)
  {
  if (InputImage[indices[i]] == 0)
    continue;		/* point was thinned */
  r=indices[i]/ncols;
  c=indices[i]%ncols;
  Transitions=0;
  if (InputImage[(r-1)*ncols+c] != L  &&  InputImage[(r-1)*ncols+c+1] == L)
    Transitions++;
  if (InputImage[(r-1)*ncols+c+1] != L  &&  InputImage[r*ncols+c+1] == L)
    Transitions++;
  if (InputImage[r*ncols+c+1] != L  &&  InputImage[(r+1)*ncols+c+1] == L)
    Transitions++;
  if (InputImage[(r+1)*ncols+c+1] != L  &&  InputImage[(r+1)*ncols+c] == L)
    Transitions++;
  if (InputImage[(r+1)*ncols+c] != L  &&  InputImage[(r+1)*ncols+c-1] == L)
    Transitions++;
  if (InputImage[(r+1)*ncols+c-1] != L  &&  InputImage[r*ncols+c-1] == L)
    Transitions++;
  if (InputImage[r*ncols+c-1] != L  &&  InputImage[(r-1)*ncols+c-1] == L)
    Transitions++;
  if (InputImage[(r-1)*ncols+c-1] != L  &&  InputImage[(r-1)*ncols+c] == L)
    Transitions++;
  if (Transitions == 1)
    {	/* can only have one edge to non-edge transition */
    endpoints[*total_ends]=r*ncols+c;
    (*total_ends)++;
    if ((*total_ends) > max_points)
      {
      MessageBox(NULL,"Error","max_points exceeded in BranchesAndEnds",MB_OK | MB_APPLMODAL);
      exit(0);
      }
    }
  else if (Transitions > 2)
    {	/* branches have > 2 edge to non-edge transitions */
    branchpoints[*total_branches]=r*ncols+c;
    (*total_branches)++;
    if ((*total_branches) > max_points)
      {
      MessageBox(NULL,"Error","max_points exceeded in BranchesAndEnds",MB_OK | MB_APPLMODAL);
      exit(0);
      }
    }
  }
}



	/*******************************************************************
	** Given an image, a starting point, and a label, this routine
	** paint-fills (8-connected) the area with the given new label.
	*******************************************************************/

int PaintFill(unsigned char	*image,
		int		ROWS,int COLS,
		int		r,int c,
		int		paint_over_label,int new_label,
		int		*queue,		/* chunk of memory to use during paint-fill */
		int		max_queue,	/* size (in cells) of queue */
		int		*indices)	/* output:  indices of pixels painted */
							/* if NULL then indices not returned */

{
int	qh,qt,count,r2,c2;

if (image[r*COLS+c] != paint_over_label)
  return(0);
image[r*COLS+c]=new_label;
if (indices != NULL)
  indices[0]=r*COLS+c;
queue[0]=r*COLS+c;
qh=1;
qt=0;
count=1;
while (qt != qh)
  {
  for (r2=-1; r2<=1; r2++)
    for (c2=-1; c2<=1; c2++)
      {
      if (r2 == 0  &&  c2 == 0)
        continue;
      if ((queue[qt]/COLS+r2) < 0  ||  (queue[qt]/COLS+r2) >= ROWS  ||
	  (queue[qt]%COLS+c2) < 0  ||  (queue[qt]%COLS+c2) >= COLS)
        continue;
      if (image[(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2] != paint_over_label)
        continue;
      image[(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2]=new_label;
	  if (indices != NULL)
        indices[count]=(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2;
      count++;
      queue[qh]=(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2;
      qh=(qh+1)%max_queue;
      if (qh == qt)
        {
        MessageBox(NULL,"Max queue size exceeded","PaintFill();",MB_OK | MB_APPLMODAL);
        exit(0);
        }
      }
  qt=(qt+1)%max_queue;
  }
return(count);
}





	/*******************************************************************
	** Traces a segment from an endpoint to its other endpoint,
	** returning the indices of the points on the segment in trace order.
	*******************************************************************/

int TraceSegment(unsigned char *image,
				 int ROWS,int COLS,
				 int r,int c,	/* starting point */
				 int *queue,int MAX_QUEUE,
				 int *indices)	/* returned indices in trace order */

{
int	qh,qt,count,r2,c2,original_label;

if (image[r*COLS+c] == 0)
  return(0);
original_label=image[r*COLS+c];
image[r*COLS+c]=1;
queue[0]=r*COLS+c;
qh=1;
qt=0;
indices[0]=r*COLS+c;
count=1;
while (qt != qh)
  {
  for (r2=-1; r2<=1; r2++)
    for (c2=-1; c2<=1; c2++)
      {
      if (r2 == 0  &&  c2 == 0)
        continue;
      if ((queue[qt]/COLS+r2) < 0  ||  (queue[qt]/COLS+r2) >= ROWS  ||
	  (queue[qt]%COLS+c2) < 0  ||  (queue[qt]%COLS+c2) >= COLS)
        continue;
      if (image[(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2] != original_label)
        continue;
      image[(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2]=1;
      indices[count]=(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2;
      count++;
      queue[qh]=(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2;
      qh=(qh+1)%MAX_QUEUE;
      if (qh == qt)
        {
        MessageBox(NULL,"TraceSegment();","MAX_QUEUE exceeded",MB_OK | MB_APPLMODAL);
        exit(0);
        }
      }
  qt=(qt+1)%MAX_QUEUE;
  }
for (r2=0; r2<ROWS*COLS; r2++)
  if (image[r2] == 1)
    image[r2]=original_label;
return(count);
}




